/* Operating Systems and Process-Oriented Programming
 * Uppsala University - 2019 Spring
 * Critical sections and mutual exclusion.
 * Demonstration of N threads created.
 * Copyright © 2019 Henrik Alfken(Schulze)
 Inspired by: https://www.it.uu.se/education/course/homepage/os/vt19/module-4/definitions/
semaphore_demo.c
 Compile and run by copy-pasting:
make bin/semaphore_demo && bin/semaphore_demo
 */
#include <pthread.h>   // pthread_create()
#include <semaphore.h> // sem_init(), sem_wait(), sem_post(), sem_destroy()
#include <stdbool.h>   // true, false
#include <stdio.h>     // printf()
#include <stdlib.h>    // abort()
#include <unistd.h>    // usleep()
#define N     3
#define SLEEP 1E6 // 1E6
#define SYNCH true     // If true synchronize, if false don't.

sem_t sem;

void *thread() {
  for (int i = 0; i < N; i++) {
    usleep(SLEEP);
    sem_post(&sem);
  }
  pthread_exit(0);
}

int main(void) {
  puts("\n - Try changing SYNCH to false and see the difference!\n");
  pthread_t tid;
  sem_init(&sem,0,0); // Second argument 0 => &sem shared between _threads_ (not processes).
  printf(" Before pthread_create: tid = %ld (uninitialized)\n",tid);
  if (pthread_create(&tid, NULL, thread, NULL) != 0) {
    perror("!ERROR! - pthread_create() failed.");
    abort();
  }
  printf("  After pthread_create: tid = %ld\n\n",tid);
  for (int i = 0; i < N; i++) {
    printf(" main(%d) waiting on a semaphore.\n", i);
    if(SYNCH){ sem_wait(&sem); } // To wait, or not to wait, that is the question. Whether...
    printf(" main(%d)   -   semaphore signaled by thread.\n", i);
  }
  sem_destroy(&sem);
  puts("\nTo wait, or not to wait, that is the question: whether 'tis nobler to be patient...");
  puts("");
  return EXIT_SUCCESS;
}
